import keysightSD1
import keysight_hvi as kthvi
import keysight_ktm5300x as ktm5300
import numpy as np

"""
Pathwave Test Sync Executive RF Sweeps using M320x AWGs M5300 RF AWGs and M9046 Chassis with High Performance Reference Clock Source (HPRCS)

Copyright Keysight Technologies 2022

In this programming example, PathWave Test Sync Executive is used to define a real-time algorithm in the FPGA (Field Programmable Gate Array) 
of Arbitrary Waveform Generators (AWGs) so that their channels can be used to output pulsed signals swept in amplitude and frequency to perform 
a spectroscopic characterization of a Device-Under-Test.

The showcased HVI API functionalities are:
* Off-shelf inter-instrument synchronization capabilities
* Scalability to an arbitrary number of instruments. In this specific case M320x AWGs and M5300 RF AWGs
* Mapping of physical quantities (amplitude, frequency, phase, etc.) into HVI registers
* Chassis interconnections and synchronization using the System Sync Module (SSM)
* Usage of M9046A Chassis with High Performance Reference Clock Source (HPRCS)

The Python code that follows: 
1. Opens an arbitrary number of M3xxxA AWGs (real hardware or simulation mode) 
2. Defines HVI resources (engines, chassis, chassis interconnections, trigger, action, events, HVI triggering resources, etc)
3. Programs HVI sequences by adding the necessary statements and instructions in each AWG's sequence
4. Compiles HVI, loads it to HW and executes it

For further info please check the programming example titled "RF Sweeps using M320x AWGs M5300 RF AWGs and M9046 Chassis with High Performance Reference Clock Source (HPRCS)"
"""

######################################
# Application Configuration 
######################################

class ApplicationConfig:
    """
    Defines module descriptors, configuration options and names of HVI engines, actions, triggers
    """
    def __init__(self):
        # Simulated HW Option
        self.hardware_simulated = True
        """
        Define PXI instruments options
        """
        # Define options to open the PXIe instruments
        # Complete option list can be found in the SD1 3.x User Guide
        self.sd1_options = 'channelNumbering=keysight'
        # Complete option list can be found in the M5300 RF AWG User Guide
        if self.hardware_simulated == True:
            self.rf_awg_instrument_options = 'Simulate=1, McapBitStreamVersionRegisterValue=1.1.146.0'
        else: 
            self.rf_awg_instrument_options = 'Simulate=0, BypassChassisCheck=1, LogLevel=Info'
        """ 
        Define names of HVI engines, actions, triggers
        """
        # Define names of HVI engines, actions, triggers
        # HVI engines
        self.rf_awg_engine_name = "RfAwgEngine"
        self.awg_engine_name = "AwgEngine"
        self.leader_engine_name = ""
        # HVI actions
        self.awg_phase_reset_name = "PhaseResetCH"
        self.awg_trigger_action_name = "AwgTrigger"
        # HVI registers
        self.freq_value_name = "Frequency Value"
        self.ampl_value_name = "Amplitude Value"
        self.freq_counter_name = "Frequency Counter"
        self.ampl_counter_name = "Amplitude Counter"
        """
        Define HW Platform
        """
        # Define module descriptors below with your instruments information
        self.module_descriptors = [
            ModuleDescriptor('M3201A', 2, 2, self.sd1_options, self.awg_engine_name),
            ModuleDescriptor('M5300', 2, 13, self.rf_awg_instrument_options, self.rf_awg_engine_name)]
        self.define_engine_names()

        # Define list of chassis containing instruments to be used by HVI
        self.chassis_list = [2]
        
        # Multi-chassis setup
        # Define the System Sync Modules included in your system.
        # For example: 
        # self.system_sync_module_descriptors = [
        #    SystemSyncModuleDescriptor('PXI0::CHASSIS2::SLOT10::INDEX0::INSTR', self.ssm_options)]
        self.ssm_options = ""
        self.ssm_simulation_options = "Simulate=True,DriverSetup=Model=M9033A"
        self.system_sync_module_descriptors = []
        # For each SSM define which SSM is connected to its downstream connectors.
        # For example: 
        # self.ssm_connections = [
        #      SystemSyncModuleConnection(ssm1_chassis=1, ssm1_downstream_connector_number=0, ssm2_chassis=2)]
        self.ssm_connections = []

        # Assign triggers to HVI object to be used for HVI-managed synchronization, data sharing, etc
        self.pxi_sync_trigger_resources = [
            kthvi.TriggerResourceId.PXI_TRIGGER0,
            kthvi.TriggerResourceId.PXI_TRIGGER1,
            kthvi.TriggerResourceId.PXI_TRIGGER2]
        """
        Define reference clock configuration
        """
        # For detailed info about all the possible configurations please read the System Setup Guide on www.keysight.com.
        # Create Ref. Clock configuration object
        self.ref_clock_config = RefClockConfig()
        # Choose the constant that reflects the ref. clock source that you are using
        self.ref_clock_source = self.ref_clock_config.SOURCE_CHASSIS_HPRCS
        # Specify the chassis number containing the ref clock source to be used
        self.ref_source_chassis_number = 2
        # Choose the ref. clock mode that you are using
        self.ref_clock_mode = self.ref_clock_config.MODE_INTERNAL
        # Specify the freq. of the signal coming form the external clock source (if connected)
        self.external_ref_frequency = 100e6 # [Hz]
        # For M5300 RF AWGs, specify if you want to connect and use the Chassis 2.4 GHz Analog Reference
        self.analog_ref_mode = self.ref_clock_config.MODE_EXTERNAL
        """
        Define system initialization
        """
        # For detailed info about system initialization please refer to KS2201A User Manual on www.keysight.com.
        # Create System Initialization configuration object
        self.sys_init_config = SystemInitializationConfig()
        # Choose the constant that reflects the system initialization step to perform
        self.sys_initialization = self.sys_init_config.MEAS_OPERATION
        """
        AWG settings
        """
        self.awg_channel = 2
        self.freq_start_value = 5e6 # [Hz]
        self.freq_step = 20e6 # [Hz]
        self.off_value = 0 # [V]
        self.ampl_start_value = 0.5 # [V]
        self.ampl_step = 0.25 # [V]
        self.pulse_duration = 200 # [ns]
        """
        RF AWG settings
        """
        self.M5xxx_fpga_clock_period = 10/3 # [ns]
        self.rf_awg_sample_rate = 4.8 # [GSa/s]
        self.waveform_file = "Waveforms/DC_m0p5_I_2048.csv"
        self.rf_awg_carrier_freq = 25.0e6
        self.rf_awg_pulse_duration = 200 # [ns]
        self.rf_awg_channel = 1 # channel playing wfm
        self.rf_awg_start_delay = 0
        self.rf_awg_num_cycles = 0 # use 0 for infinite cycles
        self.rf_awg_ampl_start_value = 30 # [%]
        self.rf_awg_ampl_step = 30 # [%]
        self.rf_awg_amplitude = 100 # [%]
        self.set_amplitude_latency = 100 # [ns]
        # Trigger settings
        self.rf_awg_trigger_mode = ktm5300.TriggerMode.SW_HVI_PER_CYCLE_TRIG
        """
        RF sweep parameters
        """
        self.num_freqs = 2
        self.num_amplitudes = 3
        self.ampl_sweep_duration = 1000 # [nS]

    def define_engine_names(self):
        """
        Define names for all HVI engines
        """
        rf_awg_count = 0
        awg_count = 0
        # Assign a number to each element of the module_descriptors list populated above
        for descriptor in self.module_descriptors:
            if descriptor.engine_name == self.awg_engine_name:
                descriptor.engine_name += str(awg_count)
                awg_count += 1
            elif descriptor.engine_name == self.rf_awg_engine_name:
                descriptor.engine_name += str(rf_awg_count)
                rf_awg_count += 1
        # Define leader engine
        self.leader_engine_name = self.module_descriptors[0].engine_name


class RefClockConfig:
    """
    Class for configuring the Ref. Clock source and modes
    """
    def __init__(self):
        # Reference clock source
        self.SOURCE_CHASSIS_REF = 0 # select this mode to use a M904x chassis internal reference as a ref. clock source
        self.SOURCE_SYSTEM_SYNC_MODULE = 1 # select this mode to use a System Sync Module (SSM) as a ref. clock source
        self.SOURCE_CHASSIS_HPRCS = 2 # select this mode to use the chassis High Performance Reference Clock Source (HPRCS) as a ref. clock source
        self.SINGLE_CHSSSIS_NO_SOURCE = 3 # select this mode to use the Chassis BackPlane (BP) in a 1-chassis system with no SSM or HPRCS
        # Internal/External clock modes
        self.MODE_INTERNAL = 0 # The ref. clock source is not locked to any external clock source
        self.MODE_EXTERNAL = 1 # The ref. clock source is locked to an external clock source


class SystemInitializationConfig:
    """
    Class for configuring the System Initialization. After putting the setup together the system initialization steps are:
    1. PRE_CALIBRATION
    2. Wait for the instruments to warmup
    3. RESET_CALIBRATION (necessary only to obtain channel skew below 50ps)
    4. MEAS_OPERATION
    Once the setup is calibrated and warmed up the normal measurement operation mode (MEAS_OPERATION) can be used.
    Consult the M5300 RF AWG User Manual and System Setup Guide for more details about the sytem initialization process.
    """
    def __init__(self):
        # System Initialization Modes
        self.PRE_CALIBRATION = 0
        self.RESET_CALIBRATION = 1
        self.MEAS_OPERATION = 2

######################################
# Main Program
######################################

def main():
    """
    Runs RF Sweeps using M3xxxA AWGs
    """
    # Load configuration
    config = ApplicationConfig()
    print()
    print("---------------------------")
    print("RF Sweeps using M320x AWGs and M5300 RF AWGs")
    print("---------------------------")
    print("AWG Frequency Sweep: Start {} MHz, Step {} MHz, {} steps, Stop {} MHz".format(config.freq_start_value/1e6, 
        config.freq_step/1e6, config.num_freqs, (config.freq_start_value+(config.num_freqs-1)*config.freq_step)/1e6))
    print("AWG Amplitude Sweep: Start Ampl. {} V, Step {} V, {} steps, Stop {} V".format(config.ampl_start_value, 
        config.ampl_step, config.num_amplitudes, config.ampl_start_value+(config.num_amplitudes-1)*config.ampl_step))
    print("RF AWG Amplitude Sweep: Start Ampl. {} %, Step {} %, {} steps, Stop {} %".format(config.rf_awg_ampl_start_value, 
        config.rf_awg_ampl_step, config.num_amplitudes, config.rf_awg_ampl_start_value+(config.num_amplitudes-1)*config.rf_awg_ampl_step))
    print()

    ###################################################
    # Open PXI modules, setup AWGs and queue waveforms
    ###################################################
    # Checks flags defining if the code executes on HW or Simulation mode
    if config.hardware_simulated:
        print("Code running in Simulation Mode")

    # Create instrument objects
    module_dict = open_modules(config)

    ######################################
    # System definition
    ######################################
    # Create system definition object
    my_system = kthvi.SystemDefinition("MySystem") 
    # Define your system, HW platform, add HVI resources
    define_hvi_resources(my_system, module_dict, config)
    # Output to console the details of the system definition
    output_sys_def(my_system, config.module_descriptors)
    
    ######################################
    # System initialization
    ######################################
    print("Initializing the defined system...")
    if config.sys_initialization == config.sys_init_config.PRE_CALIBRATION:
        my_system.initialize(kthvi.AlignmentModes.FULL | kthvi.AlignmentModes.PRE_CALIBRATION)
        print("System PRE-CALIBRATION Done")
        print("Please make sure the system is fully warmed up before moving to the next step")
        close_modules(module_dict)
        exit()
    elif config.sys_initialization == config.sys_init_config.RESET_CALIBRATION:
        my_system.initialize(kthvi.AlignmentModes.FULL | kthvi.AlignmentModes.RESET_CALIBRATION)
        print("System RESET CALIBRATION Done")
        close_modules(module_dict)
        exit()
    # Uncomment the code line below to perform a FULL Initialization that recovers the system from an unwanted state
    # my_system.initialize(kthvi.AlignmentModes.FULL)
    # Create sequencer object
    # This operation triggers a system initialization in some cases
    sequencer = kthvi.Sequencer("MySequencer", my_system)

    # Load waveforms and configures AWG channels
    # NOTE: This must be done after System Initialization
    configure_instruments(module_dict, config)

    ######################################
    # Program HVI sequences
    ######################################
    print("-----------------------")
    print("Program HVI Sequences")
    print("-----------------------")
    print("Programming the HVI sequences...")
    # Program HVI sequences using the sequencer object
    program_rf_sweeps(sequencer, module_dict, config)
    # Export the programmed sequence to text
    export_hvi_sequences(sequencer)
    
    ########################################
    # Compile, Load to HW, Run HVI Sequence
    ########################################
    print("-------------------")
    print("Execute HVI")
    print("-------------------")
    print("Compiling HVI...")
    # Compile HVI sequences
    hvi = sequencer.compile()
    print(hvi.compile_status.to_string())
    print("HVI Compiled")
    print("This HVI needs to reserve {} PXI trigger resources to execute".format(len(hvi.compile_status.sync_resources)))
    
    # Load HVI to HW: load sequences, configure actions/triggers/events, lock resources, etc.
    hvi.load_to_hw()    
    print("HVI Loaded to HW")

    # User-controlled HVI execution
    if config.hardware_simulated:
        print("Simulation completed successfully")
    else:
        interact_with_hvi(hvi, module_dict, config)

    # Release HW resources once HVI execution is completed
    print("Releasing HW...")
    hvi.release_hw()

    # Close all instrument objects
    close_modules(module_dict)

########################################################
# 2nd Level: Functions to Define, Program, Execute HVI
#######################################################

def define_hvi_resources(sys_def, module_dict, config):
    """
    Configures all the necessary resources for the HVI application to execute: HW platform, engines, actions, triggers, etc. 
    """
    # Define all the HVI engines to be included in the HVI
    define_hvi_engines(sys_def, module_dict)

    # Define list of actions to be executed
    define_hvi_actions(sys_def, module_dict, config)

    # Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
    define_hw_platform(sys_def, config)

    # Define ref. clock configuration
    define_clocking(sys_def, config)


def define_hvi_engines(sys_def, module_dict):
    """
    Define all the HVI engines to be included in the HVI
    """
    # For each instrument to be used in the HVI application add its HVI Engine to the HVI Engine Collection 
    for engine_name in module_dict.keys():
        sys_def.engines.add(module_dict[engine_name].hvi.engines.main_engine, engine_name)


def define_hvi_actions(sys_def, module_dict, config):
    """
    Define Channel Phase Reset Actions for M320x AWGs and AWG Trigger actions for M5300 RF AWGs
    """
    for engine_name, module in zip(module_dict.keys(), module_dict.values()):
        if compare_engine_type(engine_name, config.rf_awg_engine_name):
            # The phase of each M5300 RF AWG is reset by using the ABSOLUTE phase mode
            # AWG Trigger actions are necessary to trigger AWG pulses
            for ch_index in range(1, module_dict[engine_name].num_channels + 1):
                # Actions need to be added to the engine's action list so that they can be executed
                action_name = config.awg_trigger_action_name + str(ch_index) # arbitrary user-defined name
                instrument_action = "awg{}_trigger".format(ch_index) # name decided by instrument API
                action_id = getattr(module_dict[engine_name].hvi.actions, instrument_action)
                # Add AWG Trigger actions for each instrument channel
                sys_def.engines[engine_name].actions.add(action_id, action_name)
        else:
            # The phase of all M320x AWGs channels are reset synchronously from HVI
            # This way AWG channels on different instruments are ensured to be in-phase
            # For each engine, add each HVI Actions to be executed to its own HVI Action Collection
            for ch_index in range(1, module.num_channels + 1):
                # Actions need to be added to the engine's action list so that they can be executed
                # Example: hvi.engines[i].actions.add(module_dict[i].hvi.actions.ch1_reset_phase, 'AWG1_trigger')
                action_name = config.awg_phase_reset_name+ str(ch_index) # arbitrary user-defined name
                instrument_action = "ch{}_reset_phase".format(ch_index) # name decided by instrument API
                action_id = getattr(module.instrument.hvi.actions, instrument_action)
                sys_def.engines[engine_name].actions.add(action_id, action_name)


def define_hw_platform(sys_def, config):
    """
    Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks 
    """
    # Define chassis resources 
    # For multi-chassis setup details see programming example documentation
    for chassis_number in config.chassis_list:
        if config.hardware_simulated:
            sys_def.chassis.add_with_options(chassis_number, 'Simulate=True,DriverSetup=Model=GenericPxieChassis')
        else:
            sys_def.chassis.add(chassis_number)

    # Define System Sync Modules (SSMs)
    if config.system_sync_module_descriptors:
        interconnects = sys_def.interconnects
        ssm_list = []
        for descriptor in config.system_sync_module_descriptors:
            if config.hardware_simulated:
                ssm = interconnects.add_sync_module(descriptor.resource_id, config.ssm_simulation_options)
            else:
                ssm = interconnects.add_sync_module(descriptor.resource_id, descriptor.options)
            # Append SSM object
            ssm_list.append(ssm)

        # Define connections between SSMs
        if config.ssm_connections:
            for connection in config.ssm_connections:
                connector_number = connection.ssm1_downstream_connector_number
                for ssm in ssm_list:
                    if ssm.chassis == connection.ssm1_chassis:
                        ssm1 = ssm
                    if ssm.chassis == connection.ssm2_chassis:
                        ssm2 = ssm
                # Implement each user-defined connection
                try:
                    # Set connection. SSMs have always one upstream port
                    ssm1.connectivity.systemsync_downstream[connector_number].set_connection(ssm2.connectivity.systemsync_upstream[0])
                except:
                    exit("Exception! Please check the valued defined for SyncModule resource ids, chassis numbers and connections")

    # Assign the defined PXI trigger resources
    sys_def.sync_resources = config.pxi_sync_trigger_resources
    
    # Assign clock frequencies that are outside the set of the clock frequencies of each HVI engine
    # Use the code line below if you want the application to be in sync with the 10 MHz clock
    sys_def.non_hvi_core_clocks = [10e6]


def define_clocking(sys_def, config):
    """
    Define reference clock source, mode and settings.
    For detailed info about all the possible configurations please read the System Setup Guide on www.keysight.com.
    """
    # Define the Reference Clock Source
    # See System Setup Guide for detailed HW connections necessary for eac confiuguration
    if config.ref_clock_source == config.ref_clock_config.SINGLE_CHSSSIS_NO_SOURCE:
        # Chassis BackPlane (BP) is the default clock source used in a 1-chassis system that has no SSM, HPRCS or internal reference
        ref_chassis = sys_def.chassis[config.ref_source_chassis_number]
        sys_def.clocking.reference_source = ref_chassis.clock_source
    elif config.ref_clock_source == config.ref_clock_config.SOURCE_CHASSIS_HPRCS:
        # Configures High Precision Clock Reference as a source
        # It requires to connect the chassis RF Out to the SSM REF In port
        ref_chassis = sys_def.chassis[config.ref_source_chassis_number]
        hprc_source = ref_chassis.high_performance_clock_source
        sys_def.clocking.reference_source = hprc_source
    elif config.ref_clock_source == config.ref_clock_config.SOURCE_CHASSIS_REF:
        # Configures Chassis RF Analog Reference as a source. Only for M904x Chassis models
        ref_chassis = sys_def.chassis[config.ref_source_chassis_number]
        sys_def.clocking.reference_source = ref_chassis.clock_source
    elif config.ref_clock_source == config.ref_clock_config.SOURCE_SYSTEM_SYNC_MODULE:
        # Configures System Sync Module as a source
        if sys_def.interconnects:
            for index in range(sys_def.interconnects.count):
                if sys_def.interconnects[index].chassis == config.ref_source_chassis_number:
                    ssm_source = sys_def.interconnects[index]
            sys_def.clocking.reference_source = ssm_source.clock_source
        else:
            print("WARNING: Ref. Clock Source set to System Sync Module but no Sync Modules have been defined in your code! ")
            print("System clocking set to the default configuration SINGLE_CHSSSIS_NO_SOURCE")
    
    # Define the Ref. Clock Mode
    if config.ref_clock_mode == config.ref_clock_config.MODE_EXTERNAL:
        # Set up the Ext. Ref. source (if connected)
        sys_def.clocking.reference_source.set_mode(kthvi.ClockingReferenceMode.EXTERNAL, config.external_ref_frequency)
    else:
        # Default internal mode
        sys_def.clocking.reference_source.set_mode(kthvi.ClockingReferenceMode.INTERNAL)
    
    # Activate M904x Chassis Analog RF Outputs if the user connected them to M5xxx modules
    if config.analog_ref_mode == config.ref_clock_config.MODE_EXTERNAL:
        # Activate 2.4 GHz FP output for the M904x Chassis used as a reference
        clock_output_2_4GHz = ref_chassis.clock_outputs["FP2.4GHzOut"]
        clock_output_2_4GHz.set_enabled(True)
        # Control M5xxx instruments to use an external 2.4 GHz reference
        sys_def.clocking.enable_external_analog_clocks([2.4e9])


def output_sys_def(sys_def, descriptors):
    """
    Outputs the details of each instrument and engine included in the system definition.
    """
    print("-------------------")
    print("System Definition")
    print("-------------------")
    # Output KS2201A HVI Core Version
    hvi_core_version = kthvi.SystemDefinition.hvi_core_version
    print("The PathWave Test Sync Executive API installed on your system has an HVI core version " + hvi_core_version.to_string() + "\n")
    # Output info about the HW instruments
    print("HW Instruments:")
    for descriptor in descriptors:
        engine = sys_def.engines[descriptor.engine_name]
        print("- Model: {} in chassis: {}, slot: {}, HVI Engine Name: {}, HVI core version: ".format(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, descriptor.engine_name) + 
        engine.software_version.to_string() + ", HVI Engine FPGA IP version: " + engine.firmware_version.to_string())
    # Output info about the Sync Modules
    if sys_def.interconnects:
        print("System Sync Modules:")
        for index in range(sys_def.interconnects.count):
            sync_module = sys_def.interconnects[index]
            len_upstream = len(sync_module.connectivity.systemsync_upstream)
            len_downstream = len(sync_module.connectivity.systemsync_downstream)
            print("- System Sync Module in chassis: {}, slot: {} with {} Upstream Ports and {} Downstream Ports".format(sync_module.chassis, sync_module.slot, len_upstream, len_downstream))


def program_rf_sweeps(sequencer, module_dict, config):
    """ 
    Programs the nested loops for RF sweeps
    """
    # Define HVI registers
    define_registers(sequencer, module_dict, config)
    # Define sync while condition
    freq_counter = sequencer.sync_sequence.scopes[config.leader_engine_name].registers[config.freq_counter_name]
    sync_while_condition = kthvi.Condition.register_comparison(freq_counter, kthvi.ComparisonOperator.LESS_THAN, config.num_freqs)
    # Add a Sync While
    sync_while = sequencer.sync_sequence.add_sync_while("Frequency Sweep", 90, sync_while_condition)
    # Program Freq Sweep
    program_freq_sweep(sync_while.sync_sequence, module_dict, config)


def define_registers(sequencer, module_dict, config):
    """
    Defines all registers for each HVI engine
    """
    for engine_name in module_dict.keys():
        # Define registers for engines in AWG-type instruments
        if compare_engine_type(engine_name, config.awg_engine_name):
            freq_value = sequencer.sync_sequence.scopes[engine_name].registers.add(config.freq_value_name, kthvi.RegisterSize.LONG)
            freq_value.initial_value = module_dict[engine_name].instrument.freqToInt(config.freq_start_value)
            ampl_value = sequencer.sync_sequence.scopes[engine_name].registers.add(config.ampl_value_name, kthvi.RegisterSize.SHORT) 
            ampl_value.initial_value = module_dict[engine_name].instrument.voltsToInt(config.ampl_start_value)
        # Define registers for engines in RF AWG-type instruments
        elif compare_engine_type(engine_name, config.rf_awg_engine_name):
            ampl_value = sequencer.sync_sequence.scopes[engine_name].registers.add(config.ampl_value_name, kthvi.RegisterSize.SHORT) 
            ampl_value.initial_value = module_dict[engine_name].instrument.channels[0].convert_amplitude_to_int(config.rf_awg_ampl_start_value)
        # Define additional registers in the leader engine
        if engine_name==config.leader_engine_name:
            freq_counter = sequencer.sync_sequence.scopes[engine_name].registers.add(config.freq_counter_name, kthvi.RegisterSize.LONG)
            freq_counter.initial_value = 0
            ampl_counter = sequencer.sync_sequence.scopes[engine_name].registers.add(config.ampl_counter_name, kthvi.RegisterSize.SHORT)
            ampl_counter.initial_value = 0



def program_freq_sweep(sync_sequence, module_dict, config):
    """
    Programs the Frequency Sweeps
    """
    # Add Sync Multi Sequence Block
    sync_block = sync_sequence.add_sync_multi_sequence_block("Change Frequency", 800)
    # Set SMSB duration
    sync_block.duration = kthvi.time.Minimum()
    # Programs the SMSB
    program_change_freq(sync_block, module_dict, config)

    # Define sync while condition
    ampl_counter = sync_sequence.scopes[config.leader_engine_name].registers[config.ampl_counter_name]
    sync_while_condition = kthvi.Condition.register_comparison(ampl_counter, kthvi.ComparisonOperator.LESS_THAN, config.num_amplitudes)
    # Add a Sync While
    sync_while = sync_sequence.add_sync_while("Amplitude Sweep", 70, sync_while_condition)
    # Set duration of each Sync While iteration
    sync_while.duration = kthvi.time.Duration(config.ampl_sweep_duration, kthvi.time.Unit.NANOSECONDS)
    # Program Freq Sweep
    program_ampl_sweep(sync_while.sync_sequence, module_dict, config)


def program_change_freq(sync_block, module_dict, config):
    """ 
    Programs the Sync Multi-Sequence Block to Change Frequency 
    """
    # Add instructions to each local sequence within the multi-sequence block
    for engine_name in module_dict.keys():
        # Retrieve the engine sequence from the collection of HVI Local Sequences
        # HVI Local Sequence Collection is automatically created form the user-defined HVI Engine Collection
        # Each HVI Local Sequence can be retrieved using the name alias of the corresponding HVI Engine
        sequence = sync_block.sequences[engine_name]
        # Retrieve the module
        module = module_dict[engine_name]
        
        # Add instructions for AWG instruments
        if compare_engine_type(engine_name, config.awg_engine_name):
            # Execute Phase reset
            instruction = sequence.add_instruction("Reset Phase All Channels", 10, sequence.instruction_set.action_execute.id)
            instruction.set_parameter(sequence.instruction_set.action_execute.action.id, sequence.engine.actions)

            # Set Channel Frequency
            freq_value = sequence.scope.registers[config.freq_value_name]
            instruction = sequence.add_instruction("Set Channel Frequency", 200, module.hvi.instruction_set.set_frequency.id)
            instruction.set_parameter(module.hvi.instruction_set.set_frequency.channel.id, config.awg_channel)
            instruction.set_parameter(module.hvi.instruction_set.set_frequency.value.id, freq_value)

            # Set Channel Phase
            instruction = sequence.add_instruction("Set Channel Phase", 10, module.hvi.instruction_set.set_phase.id)
            instruction.set_parameter(module.hvi.instruction_set.set_phase.channel.id, config.awg_channel)
            instruction.set_parameter(module.hvi.instruction_set.set_phase.value.id, 0.0)

            # Increment freq_value
            instruction = sequence.add_instruction("Increment Frequency Value", 10, sequence.instruction_set.add.id)
            instruction.set_parameter(sequence.instruction_set.add.destination.id, freq_value)
            instruction.set_parameter(sequence.instruction_set.add.left_operand.id, freq_value)
            instruction.set_parameter(sequence.instruction_set.add.right_operand.id, module.instrument.freqToInt(config.freq_step))

            # Initialize ampl_value = ampl_start_value
            ampl_value = sequence.scope.registers[config.ampl_value_name]
            instruction = sequence.add_instruction("Initialize Amplitude Value", 10, sequence.instruction_set.assign.id)
            instruction.set_parameter(sequence.instruction_set.assign.destination.id, ampl_value)
            instruction.set_parameter(sequence.instruction_set.assign.source.id, module.instrument.voltsToInt(config.ampl_start_value))
        
        # Add instructions for RF AWG instruments
        elif compare_engine_type(engine_name, config.rf_awg_engine_name):
            # Set Channel Phase
            instruction = sequence.add_instruction("Set Channel Phase", config.M5xxx_fpga_clock_period, module.hvi.instruction_set.set_phase.id)
            instruction.set_parameter(module.hvi.instruction_set.set_phase.channel.id, config.rf_awg_channel)
            instruction.set_parameter(module.hvi.instruction_set.set_phase.value.id, 0.0)

            # Initialize ampl_value = ampl_start_value
            ampl_value = sequence.scope.registers[config.ampl_value_name]
            instruction = sequence.add_instruction("Initialize Amplitude Value", config.M5xxx_fpga_clock_period, sequence.instruction_set.assign.id)
            instruction.set_parameter(sequence.instruction_set.assign.destination.id, ampl_value)
            instruction.set_parameter(sequence.instruction_set.assign.source.id, module.instrument.channels[0].convert_amplitude_to_int(config.rf_awg_ampl_start_value))
        
        # Set and increment loop counters in the leader engine
        if engine_name == config.leader_engine_name:
            # Initialize ampl_counter = 0
            ampl_counter = sequence.scope.registers[config.ampl_counter_name]
            instruction = sequence.add_instruction("Initialize Amplitude Counter", 10, sequence.instruction_set.assign.id)
            instruction.set_parameter(sequence.instruction_set.assign.destination.id, ampl_counter)
            instruction.set_parameter(sequence.instruction_set.assign.source.id, 0)

            # Increment freq_counter
            freq_counter = sequence.scope.registers[config.freq_counter_name]
            instruction = sequence.add_instruction("Increment Frequency Counter", 10, sequence.instruction_set.add.id)
            instruction.set_parameter(sequence.instruction_set.add.destination.id, freq_counter)
            instruction.set_parameter(sequence.instruction_set.add.left_operand.id, freq_counter)
            instruction.set_parameter(sequence.instruction_set.add.right_operand.id, 1)


def program_ampl_sweep(sync_sequence, module_dict, config):
    """
    Programs the Amplitude Sweeps
    """
    # Add Sync Multi Sequence Block
    sync_block = sync_sequence.add_sync_multi_sequence_block("Change Amplitude", 260)
    # Programs the SMSB
    program_change_amplitude(sync_block, module_dict, config)


def program_change_amplitude(sync_block, module_dict, config):
    """ 
    Programs the Sync Multi-Sequence Block to Change Amplitude 
    """
    # Add instructions to each local sequence within the multi-sequence block
    for engine_name in module_dict.keys():    
        # Retrieve the engine sequence from the collection of HVI Local Sequences
        # HVI Local Sequence Collection is automatically created form the user-defined HVI Engine Collection
        # Each HVI Local Sequence can be retrieved using the name alias of the corresponding HVI Engine
        sequence = sync_block.sequences[engine_name]
        # Retrieve the module
        module = module_dict[engine_name]

        # For all modules that are RF AWG type, execute the following HVI statements
        if compare_engine_type(engine_name, config.rf_awg_engine_name):
            # Set Channel Amplitude
            ampl_value = sequence.scope.registers[config.ampl_value_name]
            instruction = sequence.add_instruction("Set Channel Amplitude", config.M5xxx_fpga_clock_period, module.hvi.instruction_set.set_amplitude.id)
            instruction.set_parameter(module.hvi.instruction_set.set_amplitude.channel.id, config.rf_awg_channel)
            instruction.set_parameter(module.hvi.instruction_set.set_amplitude.value.id, ampl_value)

            # Execute AWG Trigger action
            action_execute_instr = sequence.instruction_set.action_execute
            instr = sequence.add_instruction("AWG Trigger", config.set_amplitude_latency, action_execute_instr.id)
            instr.set_parameter(action_execute_instr.action.id, sequence.engine.actions[config.awg_trigger_action_name+str(1)])

            # Increment ampl_value
            instruction = sequence.add_instruction("Increment Amplitude Value", config.M5xxx_fpga_clock_period, sequence.instruction_set.add.id)
            instruction.set_parameter(sequence.instruction_set.add.destination.id, ampl_value)
            instruction.set_parameter(sequence.instruction_set.add.left_operand.id, ampl_value)
            instruction.set_parameter(sequence.instruction_set.add.right_operand.id, module.instrument.channels[0].convert_amplitude_to_int(config.rf_awg_ampl_step))

        # For all modules that are AWG type, execute the following HVI statements
        else:
            # Set Channel Amplitude ON
            ampl_value = sequence.scope.registers[config.ampl_value_name]
            instruction = sequence.add_instruction("Set Channel Amplitude: ON", 10, module.hvi.instruction_set.set_amplitude.id)
            instruction.set_parameter(module.hvi.instruction_set.set_amplitude.channel.id, config.awg_channel)
            instruction.set_parameter(module.hvi.instruction_set.set_amplitude.value.id, ampl_value)
            # Set Channel Amplitude OFF
            instruction = sequence.add_instruction("Set Channel Amplitude: OFF", config.pulse_duration, module.hvi.instruction_set.set_amplitude.id)
            instruction.set_parameter(module.hvi.instruction_set.set_amplitude.channel.id, config.awg_channel)
            instruction.set_parameter(module.hvi.instruction_set.set_amplitude.value.id, config.off_value)
            # Increment ampl_value
            instruction = sequence.add_instruction("Increment Amplitude Value", 10, sequence.instruction_set.add.id)
            instruction.set_parameter(sequence.instruction_set.add.destination.id, ampl_value)
            instruction.set_parameter(sequence.instruction_set.add.left_operand.id, ampl_value)
            instruction.set_parameter(sequence.instruction_set.add.right_operand.id, module.instrument.voltsToInt(config.ampl_step))

        # Increment loop counter in the leader engine
        if engine_name == config.leader_engine_name:
            # Increment ampl_counter register value
            ampl_counter = sequence.scope.registers[config.ampl_counter_name]
            instruction = sequence.add_instruction("Increment Amplitude Counter", 10, sequence.instruction_set.add.id)
            instruction.set_parameter(sequence.instruction_set.add.destination.id, ampl_counter)
            instruction.set_parameter(sequence.instruction_set.add.left_operand.id, ampl_counter)
            instruction.set_parameter(sequence.instruction_set.add.right_operand.id, 1)
            # Add a delay to make sure the register value gets update before the next iteration
            sequence.add_delay("Loop Delay", 50)


def compare_engine_type(engine_name, instrument_name):
    """
    This function compares HVI engine names to determine if an engine belongs to a certain instrument type .
    The assumption is that engine naming is assigned as InstrumentTypeK, K = 0, 1, 2, 3..
    """
    # Load all possible instrument types from ApplicationConfig class
    config = ApplicationConfig()
    min_length = min(len(engine_name), len(config.awg_engine_name), len(config.rf_awg_engine_name))

    if engine_name[:min_length] == instrument_name[:min_length]:
        return True
    else:
        return False


def export_hvi_sequences(sequencer):
    """
    Exports the programmed HVI sequences to text format
    """
    # Generate HVI sequence description text
    output = sequencer.sync_sequence.to_string(kthvi.OutputFormat.DEBUG)
    print("Programmed HVI sequences exported to file")
    
    # Write file
    text_file = open("TextOutput/HviRfSweeps.txt", "w")
    text_file.write(output)
    text_file.close()


def interact_with_hvi(hvi, module_dict, config):
    """
    User-controlled HVI execution 
    """
    # If running with hardware, iteratively re-run HVI until the user decides to exit
    while True:
        # Execute HVI in blocking mode: SW waits until HVI sequences ends their execution
        # Eventually enter a timeout for the HVI execution to be stopped: timeout = timedelta(seconds=0), hvi.run(timeout)
        hvi.run(hvi.no_timeout)
        print("HVI Running...")
        # Print registers values
        leader_module = module_dict[config.leader_engine_name]
        leader_registers = hvi.sync_sequence.scopes[config.leader_engine_name].registers
        freq_value = leader_registers[config.freq_value_name].read() * (config.freq_step/leader_module.instrument.freqToInt(config.freq_step))
        ampl_value = leader_registers[config.ampl_value_name].read() * (config.ampl_step/leader_module.instrument.voltsToInt(config.ampl_step))
        # Note: The register containing the amplitude, frequency values are incremented once more before exiting the loop but the values are not applied
        print("Final Register Values: Amplitude {} V, Frequency {} MHz".format(ampl_value, freq_value/1e6))
        print("Press enter to run HVI again, q to exit...")
        if input() == 'q':
            break

#################################################################
# 3rd Level: Classes and Functions to Use SD1/M3xxxA Instruments
#################################################################

class ModuleDescriptor:
    """
    Descriptor for module objects
    """
    def __init__(self, model_number, chassis_number, slot_number, options, engine_name):
        self.model_number = model_number
        self.chassis_number = chassis_number
        self.slot_number = slot_number
        self.options = options
        self.engine_name = engine_name


class SystemSyncModuleDescriptor:
    """
    Descriptor for System Sync Modules
    """
    def __init__(self, resource_id, options):
        self.resource_id = resource_id
        self.options = options


class SystemSyncModuleConnection:
    """
    Describes the connection between two SyncModule objects ssm1 and ssm2:
    ssm1.connectivity.systemsync_downstream[ssm1_downstream_connector_number].connection = ssm2.connectivity.systemsync_upstream[1]
    """
    def __init__(self, ssm1_chassis, ssm1_downstream_connector_number, ssm2_chassis):
        self.ssm1_chassis = ssm1_chassis
        self.ssm1_downstream_connector_number = ssm1_downstream_connector_number
        self.ssm2_chassis = ssm2_chassis


class Module:
    """
    Class defining a modular instrument object and its properties
    """
    def __init__(self, instrument_object, hvi_interface, num_channels):
        self.instrument = instrument_object
        self.hvi = hvi_interface
        self.num_channels = num_channels


def open_modules(config):
    """ 
    Opens and creates each PXIe instrument object.
    Returns a dictionary of module objects whose keys are the HVI engine names.
    Please check each instrument User Guide for options to open the instrument objects.
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules
    # Query and reset options
    query = True
    reset = True

    # Open PXIe instrument objects
    for descriptor in config.module_descriptors:
        # Open M5300 RF AWGs
        if descriptor.model_number == 'M5300':
            resource_id = f"PXI0::CHASSIS{descriptor.chassis_number}::SLOT{descriptor.slot_number}::INSTR"
            instr_obj_options =  descriptor.options
            instr_obj = ktm5300.KtM5300x(resource_id, query, reset, instr_obj_options)
            # Create HVI interface
            hvi = ktm5300.KtM5300xHvi(instr_obj.get_instance())
            # M5xxx RF AWGs have 4 channels
            num_channels = 4
        elif descriptor.model_number[:2] == 'M3':
            # Open SD1 instrument objects
            if descriptor.model_number == "M3102A":
                instr_obj = keysightSD1.SD_AIN()
            else:
                instr_obj = keysightSD1.SD_AOU()
            instr_obj_options = descriptor.options + ',simulate=true' if config.hardware_simulated else descriptor.options
            id = instr_obj.openWithOptions(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, instr_obj_options)
            if id < 0:
                raise Exception("Error opening instrument in chassis: {}, slot: {}! Error code: {} - {}. Exiting...".format(descriptor.chassis_number, descriptor.slot_number, id, keysightSD1.SD_Error.getErrorMessage(id)))
            # Create HVI interface
            hvi = instr_obj.hvi
            # Get number of channels
            nCh = instr_obj.getOptions("channels")
            if nCh == "CH2":
                num_channels = 2
            elif nCh == "CH4":
                num_channels = 4
            else:
                raise Exception("PXI module in chassis {}, slot {} returned number of channels = {} which is incorrect. Exiting... ".format(instr_obj.getChassis(), instr_obj.getSlot(), nCh))
        else:
            raise Exception("The descriptor for PXI module in chassis {}, slot {} has a model number {} which is not supported by this prog. example. Exiting... ".format(descriptor.chassis_number, descriptor.slot_number, descriptor.model_number))
        # Populate module dictionary
        module_dict[descriptor.engine_name] = Module(instr_obj, hvi, num_channels)
        num_modules += 1
    
    return module_dict


def configure_instruments(module_dict, config):
    """
    Configures each PXIe instrument
    """
    for descriptor in config.module_descriptors:
        module = module_dict[descriptor.engine_name]
        if descriptor.model_number == 'M5300':
            # Configure each RF AWG and queue waveforms
            configure_rf_awg(config, module)
        elif descriptor.model_number[:3] == 'M32':
            # Configure each M320x AWG
            configure_awg(config, module)
        else:
            raise Exception(f"The descriptor for PXI module in chassis {descriptor.chassis_number}," + 
                            f"slot {descriptor.slot_number} has a model number {descriptor.model_number}" + 
                            "which is not supported by this prog. example. Exiting... ")


def configure_awg(config, awg_module, num_channels = 2):
    """ 
    Configures M320x AWG channels to be used to output RF pulses
    """
    awg_instrument = awg_module.instrument
    # Configure mode, amlitude, offset of every channel
    for channel in range(1, num_channels+1):
        awg_instrument.channelWaveShape(channel, keysightSD1.SD_Waveshapes.AOU_SINUSOIDAL)
        awg_instrument.channelAmplitude(channel, config.off_value)
        awg_instrument.channelOffset(channel, 0)
        awg_instrument.channelPhase(channel, 0)


def configure_rf_awg(config, rf_awg_module):
    """ 
    Configures M5300 RF AWG channels to be used to output modulated RF pulses
    """
    rf_awg_instrument = rf_awg_module.instrument
    num_channels = rf_awg_module.num_channels
    # Define arbitrary waveform
    length = int(config.rf_awg_pulse_duration * config.rf_awg_sample_rate)
    wfm = np.ones(length, dtype=np.double)
    # Configure arbitrary waveform settings
    arb_wfm = rf_awg_instrument.channels[0].arbitrary_waveform
    arb_wfm.clear_waveform_memory() # clears AWG memory
    wfm_id = 1
    # To create waveform from file use ti scode instead:
    # wfm_file = config.waveform_file
    # arb_wfm.create_from_file(wfm_file, wfm_id, ktm5300.WaveformFileType.ANALOG)
    arb_wfm.create(wfm, wfm_id)
    arb_wfm.burst_count = 1
    arb_wfm.padding_mode = ktm5300.PaddingMode.DEFAULT_SAMPLE
    arb_wfm.playback_mode = ktm5300.PlaybackMode.ONE_SHOT
    arb_wfm.queue_output_on_stop = ktm5300.QueueOutputOnStop.GO_TO0_V

    # Configure every AWG channel
    # NOTE: M5xxx channel numbering starts from 0
    for ch in range(num_channels):
        channel = rf_awg_instrument.channels[ch]
        channel.enabled = True
        # AWG queue flush 
        channel.awg_stop()
        channel.awg_queue_flush()
        # Set amplitude, phase
        channel.amplitude = config.rf_awg_amplitude
        channel.phase = 0
        channel.phase_mode = ktm5300.PhaseMode.ABSOLUTE
        # Configure AWG settings
        channel.upconverter.local_oscillator_frequency = config.rf_awg_carrier_freq
        channel.output_type = ktm5300.OutputWaveform.MODULATED
        # Queue waveform to the selected channel
        channel.awg_queue_waveform(config.rf_awg_trigger_mode, config.rf_awg_start_delay, config.rf_awg_num_cycles, wfm_id)
        channel.awg_start() # AWG starts and waits for an AWG trigger


def close_modules(module_dict):
    """
    Close all modules at the end of the execution
    """
    for engine_name in module_dict:
        module_dict[engine_name].instrument.close()
    print("PXI modules closed\n")

if __name__ == '__main__':
    main()